# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for parsing xry files."""
from gettext import pgettext as C_

from gi.repository import Graphs

from graphs.file_import.parsers import Parser


class XryParser(Parser):
    """Xry parser."""

    __gtype_name__ = "GraphsPythonXryParser"

    def __init__(self):
        super().__init__(
            "xry",
            C_("import-mode", "xry"),
            C_("file-filter", "Leybold xry"),
            ["xry"],
        )

    @staticmethod
    def parse(
        items: Graphs.ItemList,
        settings: Graphs.ImportSettings,
        data: Graphs.Data,
    ) -> None:
        """Import data from .xry files used by Leybold X-ray apparatus."""
        parser = Graphs.XryParser.new()
        parser.parse(data, settings, items)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for parsing spreadsheet files (ODS/XLSX)."""
import xml.etree.ElementTree
import zipfile
from gettext import gettext as _
from gettext import pgettext as C_

from gi.repository import GLib, Gio, Graphs, Gtk

from graphs import file_io
from graphs.file_import.parsers import Parser
from graphs.item import DataItem
from graphs.misc import ParseError

import numexpr

import numpy

# XML Namespaces
ODS_TABLE_NAMESPACE = "urn:oasis:names:tc:opendocument:xmlns:table:1.0"
XLSX_MAIN_NAMESPACE = \
    "http://schemas.openxmlformats.org/spreadsheetml/2006/main"


class OdsParser:
    """ODS file parser."""

    def get_sheet_names(self, file: Gio.File) -> list[str]:
        """Get sheet names from ODS file."""
        with file_io.open(file, "rb") as file_obj, \
             zipfile.ZipFile(file_obj) as zip_file, \
             zip_file.open("content.xml") as content_file:
            root = xml.etree.ElementTree.parse(content_file).getroot()
            namespaces = {"table": ODS_TABLE_NAMESPACE}
            sheets = root.findall(".//table:table", namespaces)
            attribute_name = f"{{{namespaces['table']}}}name"
            return [sheet.get(attribute_name) for sheet in sheets]

    def parse_file(
        self,
        file: Gio.File,
        columns: set[int],
        sheet_index: int,
    ) -> list[tuple[str, list[float]]]:
        """Parse ODS file and return list of requested columns."""
        with file_io.open(file, "rb") as file_obj, \
             zipfile.ZipFile(file_obj) as zip_file, \
             zip_file.open("content.xml") as content_file:
            root = xml.etree.ElementTree.parse(content_file).getroot()
            namespaces = {"table": ODS_TABLE_NAMESPACE}
            repeat_key = f"{{{namespaces['table']}}}number-columns-repeated"
            sheets = root.findall(".//table:table", namespaces)

            max_col = max(columns)
            raw_columns = {col_index: [[], ""] for col_index in columns}

            table_rows = sheets[sheet_index].findall(
                "table:table-row",
                namespaces,
            )
            for table_row in table_rows:
                cells = table_row.findall("table:table-cell", namespaces)
                current_col = 0

                for table_cell in cells:
                    repeat_count = int(table_cell.get(repeat_key, 1))
                    cell_value = "".join(table_cell.itertext())
                    for _count in range(repeat_count):
                        if current_col > max_col:
                            break
                        if current_col in columns:
                            try:
                                val = Graphs.evaluate_string(cell_value)
                                raw_columns[current_col][0].append(val)
                            except GLib.Error:
                                if len(raw_columns[current_col][0]) == 0:
                                    raw_columns[current_col][1] = \
                                        cell_value.strip()
                                else:
                                    break
                        current_col += 1
            return raw_columns


class XlsxParser:
    """XLSX file parser."""

    def get_sheet_names(self, file: Gio.File) -> list[str]:
        """Get sheet names from XLSX file."""
        with file_io.open(file, "rb") as file_obj, \
             zipfile.ZipFile(file_obj) as zip_file, \
             zip_file.open("xl/workbook.xml") as workbook_file:
            root = xml.etree.ElementTree.parse(workbook_file).getroot()
            namespaces = {"main": XLSX_MAIN_NAMESPACE}
            sheets = root.findall(f".//{{{namespaces['main']}}}sheet")
            return [sheet.get("name") for sheet in sheets]

    def parse_file(
        self,
        file: Gio.File,
        columns: set[int],
        sheet_index: int,
    ) -> list[tuple[str, list[float]]]:
        """Parse XLSX file and return 2D array of cell values."""
        with file_io.open(file, "rb") as file_obj, \
             zipfile.ZipFile(file_obj) as zip_file:
            shared_strings = self._load_shared_strings(zip_file)
            return self._parse_worksheet(
                zip_file,
                columns,
                sheet_index,
                shared_strings,
            )

    def _load_shared_strings(self, zip_file: zipfile.ZipFile) -> list[str]:
        """Load shared strings from XLSX file."""
        try:
            with zip_file.open("xl/sharedStrings.xml") as strings_file:
                root = xml.etree.ElementTree.parse(strings_file).getroot()
                namespace = XLSX_MAIN_NAMESPACE
                text_elements = root.findall(f".//{{{namespace}}}t")
                return [element.text or "" for element in text_elements]
        except KeyError:
            return []  # No shared strings in this file

    def _parse_worksheet(
        self,
        zip_file: zipfile.ZipFile,
        columns: set[int],
        sheet_index: int,
        shared_strings: list[str],
    ) -> list[list[str]]:
        """Parse worksheet and return requested columns."""
        sheet_file = f"xl/worksheets/sheet{sheet_index + 1}.xml"
        with zip_file.open(sheet_file) as worksheet_file:
            root = xml.etree.ElementTree.parse(worksheet_file).getroot()
            namespaces = {"main": XLSX_MAIN_NAMESPACE}
            raw_columns = {col_index: [[], ""] for col_index in columns}
            for row_element in root.findall(".//main:row", namespaces):
                row_data = self._parse_row(
                    row_element,
                    namespaces,
                    shared_strings,
                    columns,
                )
                for col_index in columns:
                    cell_value = row_data.get(col_index, "")
                    try:
                        val = Graphs.evaluate_string(cell_value)
                        raw_columns[col_index][0].append(val)
                    except GLib.Error:
                        if len(raw_columns[col_index][0]) == 0:
                            raw_columns[col_index][1] = cell_value.strip()
                        else:
                            break
        return raw_columns

    def _parse_row(
        self,
        row_element,
        namespaces: dict,
        shared_strings: list[str],
        columns: set,
    ) -> dict:
        """Parse a single row element, returning only requested columns."""
        cell_elements = row_element.findall("main:c", namespaces)
        cell_data = {}

        for cell_element in cell_elements:
            ref = cell_element.get("r")
            column = "".join(c for c in ref if not c.isdigit())
            column_index = Graphs.tools_alpha_to_int(column)

            if column_index in columns:
                cell_data[column_index] = self._get_cell_value(
                    cell_element,
                    namespaces,
                    shared_strings,
                )
        return cell_data

    def _get_cell_value(
        self,
        cell_element,
        namespaces: dict,
        shared_strings: list[str],
    ) -> str:
        """Extract cell value, handling shared strings."""
        value = cell_element.find("main:v", namespaces)
        cell_string = value.text if value is not None else ""
        cell_type = cell_element.get("t")

        if cell_type == "s":  # Type s is for shared strings
            string_index = int(cell_string)
            cell_string = shared_strings[string_index]

        return cell_string


class SpreadsheetParser(Parser):
    """Main spreadsheet parser that delegates to ODS/XLSX parsers."""

    def __init__(self):
        super().__init__(
            "spreadsheet",
            C_("import-mode", "Spreadsheet"),
            C_("file-filter", "Spreadsheet"),
            ["ods", "xlsx"],
        )

    @staticmethod
    def init_settings(settings: Graphs.ImportSettings) -> bool:
        """Init settings with default spreadsheet sheet selection."""
        file = settings.get_file()
        parser = \
            OdsParser() if file.get_path().endswith(".ods") else XlsxParser()
        try:
            sheet_names = parser.get_sheet_names(file)
            string_list = Gtk.StringList.new(sheet_names)
            settings.set_item("sheet-names", string_list)
            return True
        except zipfile.BadZipFile:
            return False

    @staticmethod
    def init_settings_widgets(
        settings: Graphs.ImportSettings,
        box: Gtk.Box,
    ) -> None:
        """Append Spreadsheet-specific settings widgets."""
        if not settings.get_item("sheet-names"):
            return
        box.append(Graphs.SpreadsheetGroup.new(settings))
        box.append(Graphs.SpreadsheetBox.new(settings))

    @staticmethod
    def parse(
        items: Graphs.ItemList,
        settings: Graphs.ImportSettings,
        data: Graphs.Data,
    ) -> None:
        """Import data from ODS or XLSX file."""
        style = data.get_selected_style_params()
        file = settings.get_file()
        sheet_index = settings.get_int("sheet-index")

        column_indices = set()
        item_settings_list = []
        variant = settings.get_value("items")
        for i in range(variant.n_children()):
            item_settings = Graphs.ColumnsItemSettings()
            item_settings.load_from_variant(variant.get_child_value(i))
            item_settings_list.append(item_settings)

            column_indices.add(item_settings.column_y)
            if not item_settings.single_column:
                column_indices.add(item_settings.column_x)
            if item_settings.use_xerr:
                column_indices.add(item_settings.xerr_index)
            if item_settings.use_yerr:
                column_indices.add(item_settings.yerr_index)

        file_parser = \
            OdsParser() if file.get_path().endswith(".ods") else XlsxParser()
        parsed_columns = file_parser.parse_file(
            file,
            column_indices,
            sheet_index,
        )

        for item_settings in item_settings_list:
            ydata, ylabel = parsed_columns[item_settings.column_y]

            if len(ydata) == 0:
                raise ParseError(_("No numeric data found in column."))

            if item_settings.single_column:
                xlabel = ""
                xdata = numexpr.evaluate(
                    Graphs.preprocess_equation(item_settings.equation),
                    local_dict={"n": numpy.arange(len(ydata))},
                )
                if xdata.ndim == 0:
                    xdata = numpy.full(len(ydata), xdata)
            else:
                xdata, xlabel = parsed_columns[item_settings.column_x]
                if len(xdata) != len(ydata):
                    raise ParseError(_("Columns do not have the same length."))

            xerr, _label = parsed_columns[item_settings.xerr_index] \
                if item_settings.use_xerr else (None, None)
            yerr, _label = parsed_columns[item_settings.yerr_index] \
                if item_settings.use_yerr else (None, None)

            items.add(
                DataItem.new(
                    style,
                    xdata,
                    ydata,
                    xerr=xerr,
                    yerr=yerr,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    name=settings.get_filename(),
                ),
            )



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for importing data from files."""
import logging
from gettext import gettext as _
from pathlib import Path

from gi.repository import Graphs, Gtk

from graphs.file_import import parsers
from graphs.file_import.parsers import (
    columns,
    project,
    spreadsheet,
    sql,
    xrdml,
    xry,
)
from graphs.misc import ParseError

_REQUESTS = (
    "guess_import_mode",
    "init_import_settings",
    "append_settings_widgets",
    "parse",
)


class DataImporter(Graphs.DataImporter):
    """Class for importing data."""

    __gtype_name__ = "GraphsPythonDataImporter"

    def __init__(self):
        super().__init__()
        for request in _REQUESTS:
            request = request + "-request"
            self.connect(
                request,
                getattr(self, "_on_" + request.replace("-", "_")),
            )

        # Order determines UI order
        parsers.register_parser(columns.ColumnsParser())
        parsers.register_parser(project.ProjectParser())
        parsers.register_parser(sql.SqlParser())
        parsers.register_parser(spreadsheet.SpreadsheetParser())
        parsers.register_parser(xrdml.XrdmlParser())
        parsers.register_parser(xry.XryParser())

        self.setup(parsers.list_parsers())

    @staticmethod
    def _on_guess_import_mode_request(
        self,
        settings: Graphs.ImportSettings,
    ) -> int:
        """
        Guess the import mode of a file.

        At the moment this relies purely on the file extension.
        """
        try:
            filename = Graphs.tools_get_filename(settings.get_file())
            file_suffix = Path(filename).suffixes[-1][1:]
        except IndexError:
            file_suffix = None
        for index, mode in enumerate(parsers.list_parsers()):
            suffixes = mode.get_file_suffixes()
            if suffixes is None:
                continue
            for suffix in suffixes:
                if file_suffix == suffix:
                    return index
        return 0  # columns

    @staticmethod
    def _on_init_import_settings_request(
        self,
        settings: Graphs.ImportSettings,
    ) -> None:
        """
        Init import settings.

        This is intended for filetypes, where the settings are dependent on
        the file itself or other runtime variables.
        """
        parser = parsers.get_parser(settings.get_mode())
        try:
            return parser.init_settings(settings)
        except Exception:
            return False

    @staticmethod
    def _on_append_settings_widgets_request(
        self,
        settings: Graphs.ImportSettings,
        settings_box: Gtk.Box,
    ) -> Gtk.Widget:
        """Load the UI settings."""
        parser = parsers.get_parser(settings.get_mode())
        parser.init_settings_widgets(settings, settings_box)

    @staticmethod
    def _on_parse_request(
        self,
        itemlist: Graphs.ItemList,
        settings: Graphs.ImportSettings,
        data: Graphs.Data,
    ) -> str:
        parser = parsers.get_parser(settings.get_mode())
        try:
            parser.parse(itemlist, settings, data)
            return ""
        except ParseError as error:
            return error.message
        except Exception:
            message = _("Import failed")
            logging.exception(message)
            return message



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for Exporting data."""
import sys

from gi.repository import Gio, Graphs

from graphs import utilities


def export_items(
    mode: str,
    file: Gio.File,
    items: list[Graphs.Item],
    figure_settings: Graphs.FigureSettings,
) -> None:
    """Export items in specified format."""
    callback = getattr(sys.modules[__name__], "_export_" + mode)
    callback(file, items, figure_settings)


def _export_columns(
    file: Gio.File,
    items: list[Graphs.Item],
    figure_settings: Graphs.FigureSettings,
) -> None:
    """Save Items in columns format."""
    if len(items) > 1:
        for item in items:
            name = f"{item.get_name()}.txt"
            _save_item(
                file.get_child_for_display_name(name),
                item,
                figure_settings,
            )
    else:
        _save_item(file, items[0], figure_settings)


def _save_item(
    file: Gio.File,
    item: Graphs.DataItem | Graphs.EquationItem | Graphs.GeneratedDataItem,
    figure_settings: Graphs.FigureSettings,
) -> None:
    """Save Item in columns format."""
    delimiter = "\t"
    xlabel, ylabel = item.get_xlabel(), item.get_ylabel()
    stream = Gio.DataOutputStream.new(
        file.replace(None, False, Gio.FileCreateFlags.NONE, None),
    )

    xerr, yerr = None, None
    if isinstance(item, (Graphs.DataItem, Graphs.GeneratedDataItem)):
        xdata, ydata, xerr, yerr = item.get_data_tuple()
    elif isinstance(item, Graphs.EquationItem):
        limits = figure_settings.get_limits().values()
        if item.get_xposition() == Graphs.XPosition.BOTTOM:
            limits = [limits[0], limits[1]]
        else:
            limits = [limits[2], limits[3]]
        equation = item.get_preprocessed_equation()
        xdata, ydata = utilities.equation_to_data(equation, limits)

    n_cols = 2 + (xerr is not None) + (yerr is not None)
    fmt = delimiter.join(["%.12e"] * n_cols)

    if xlabel != "" and ylabel != "":
        headers = [xlabel, ylabel]
        if xerr is not None:
            headers.append("x_err")
        if yerr is not None:
            headers.append("y_err")
        stream.put_string(delimiter.join(headers) + "\n")

    err_cols = [e for e in (xerr, yerr) if e is not None]
    for values in zip(xdata, ydata, *err_cols):
        stream.put_string(fmt % values + "\n")
    stream.close()



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for data Items."""
from gi.repository import GObject, Graphs

from graphs import misc, utilities

from matplotlib import RcParams

import numpy


class _PythonItemMixin:

    def reset(
        self,
        old_style: tuple[RcParams, dict],
        new_style: tuple[RcParams, dict],
    ) -> None:
        """Reset all properties."""
        if not hasattr(self, "_style_properties"):
            return
        # Combine rcparams and graphs_params into single dict:
        old_style = old_style[0] | old_style[1]
        new_style = new_style[0] | new_style[1]
        for prop, (key, function) in self._style_properties.items():
            old_value = old_style[key]
            new_value = new_style[key]
            if function is not None:
                old_value = function(old_value)
                new_value = function(new_value)
            if self.get_property(prop) == old_value:
                self.set_property(prop, new_value)

    def override(
        self,
        style: tuple[RcParams, dict],
    ) -> None:
        """Override all properties."""
        if not hasattr(self, "_style_properties"):
            return
        # Combine rcparams and graphs_params into single dict:
        style = style[0] | style[1]
        for prop, (key, function) in self._style_properties.items():
            value = style[key] if function is None else function(style[key])
            self.set_property(prop, value)

    @staticmethod
    def _extract_params(
        cls,
        style: tuple[RcParams, dict],
        kwargs: dict,
    ) -> dict:
        style = style[0] | style[1]  # Add graphs_params to style dict
        return {
            prop: style[key] if function is None else function(style[key])
            for prop, (key, function) in cls._style_properties.items()
            if prop not in kwargs
        }

    def to_dict(self) -> dict:
        """Convert item to dict."""
        dictionary = {
            key: self.get_property(key)
            for key in dir(self.props) if key != "typename"
        }
        dictionary["type"] = self.__gtype_name__[12:]
        return dictionary


class DataItem(Graphs.DataItem, _PythonItemMixin):
    """DataItem."""

    __gtype_name__ = "GraphsPythonDataItem"

    _style_properties = {
        "errbarsabove": ("errorbar.barsabove", None),
        "errcapsize": ("errorbar.capsize", None),
        "errcapthick": ("errorbar.capthick", None),
        "errlinewidth": ("errorbar.linewidth", None),
        "linestyle": ("lines.linestyle", misc.LINESTYLES.index),
        "linewidth": ("lines.linewidth", None),
        "markerstyle": ("lines.marker", misc.MARKERSTYLES.index),
        "markersize": ("lines.markersize", None),
    }

    @classmethod
    def new(
        cls,
        style: tuple[RcParams, dict],
        xdata: list[float] = None,
        ydata: list[float] = None,
        xerr: list[float] = None,
        yerr: list[float] = None,
        **kwargs,
    ):
        """Create new DataItem."""
        data = Graphs.DataHolder.new(xdata, ydata, xerr, yerr)
        return cls.new_with_data(style, data, **kwargs)

    @classmethod
    def new_with_data(
        cls,
        style: tuple[RcParams, dict],
        data: Graphs.DataHolder,
        **kwargs,
    ):
        """Create new DataItem with a DataHolder."""
        return cls(
            data=data,
            **cls._extract_params(cls, style, kwargs),
            **kwargs,
        )

    def to_dict(self) -> dict:
        """Convert item to dict."""
        dictionary = super().to_dict()
        dictionary["data"] = self.get_data_tuple()
        return dictionary

    def get_data_tuple(self) -> tuple[list, list, list, list]:
        """Get the data as a picklable tuple."""
        return (
            self.get_xdata().tolist(),
            self.get_ydata().tolist(),
            self.get_xerr().tolist() if self.has_xerr() else None,
            self.get_yerr().tolist() if self.has_yerr() else None,
        )

    def set_data_tuple(self, data: tuple[list, list, list, list]) -> None:
        """Set the data from a tuple."""
        self.props.data = Graphs.DataHolder.new(*data)

    def get_xydata(self) -> tuple[numpy.ndarray, numpy.ndarray]:
        """Get x- and y-data."""
        return self.get_xdata(), self.get_ydata()

    def set_xydata(self, xydata: tuple[numpy.ndarray, numpy.ndarray]) -> None:
        """Set x- and y-data."""
        self.set_data_tuple((*xydata, self.get_xerr(), self.get_yerr()))

    def get_xdata(self) -> numpy.ndarray:
        """Get xdata."""
        return utilities.bytes_to_ndarray(self.props.data.get_xdata_b())

    def get_ydata(self) -> numpy.ndarray:
        """Get ydata."""
        return utilities.bytes_to_ndarray(self.props.data.get_ydata_b())

    def get_xerr(self) -> numpy.ndarray:
        """Get xerr."""
        return utilities.bytes_to_ndarray(self.props.data.get_xerr_b())

    def get_yerr(self) -> numpy.ndarray:
        """Get yerr."""
        return utilities.bytes_to_ndarray(self.props.data.get_yerr_b())


class GeneratedDataItem(Graphs.GeneratedDataItem, DataItem):
    """Generated Dataitem."""

    __gtype_name__ = "GraphsPythonGeneratedDataItem"

    @classmethod
    def new(
        cls,
        style: tuple[RcParams, dict],
        equation: str,
        xstart: str,
        xstop: str,
        steps: int,
        scale: Graphs.Scale,
        **kwargs,
    ):
        """Create new GeneratedDataItem."""
        return cls(
            equation=equation,
            xstart=xstart,
            xstop=xstop,
            steps=steps,
            scale=scale,
            **cls._extract_params(cls, style, kwargs),
            **kwargs,
        )


class EquationItem(Graphs.EquationItem, _PythonItemMixin):
    """EquationItem."""

    __gtype_name__ = "GraphsPythonEquationItem"

    _style_properties = {
        "linestyle": (
            "lines.linestyle",
            lambda x: max(misc.LINESTYLES.index(x) - 1, 0),
        ),
        "linewidth": ("lines.linewidth", None),
    }

    @classmethod
    def new(cls, style: tuple[RcParams, dict], equation: str, **kwargs):
        """Create new EquationItem."""
        return cls(
            equation=equation,
            **cls._extract_params(cls, style, kwargs),
            **kwargs,
        )


class TextItem(Graphs.TextItem, _PythonItemMixin):
    """TextItem."""

    __gtype_name__ = "GraphsPythonTextItem"

    _style_properties = {
        "size": ("font.size", None),
        "color": ("text.color", None),
    }

    @classmethod
    def new(
        cls,
        style: tuple[RcParams, dict],
        xanchor: float = 0,
        yanchor: float = 0,
        text: str = "",
        **kwargs,
    ):
        """Create new textItem."""
        return cls(
            xanchor=xanchor,
            yanchor=yanchor,
            text=text,
            **cls._extract_params(cls, style, kwargs),
            **kwargs,
        )


class FillItem(Graphs.FillItem, _PythonItemMixin):
    """FillItem."""

    __gtype_name__ = "GraphsPythonFillItem"

    data = GObject.Property(type=object)

    @classmethod
    def new(
        cls,
        _params: tuple[RcParams, dict],
        data: tuple[list[float], list[float], list[float]],
        **kwargs,
    ):
        """Create new FillItem."""
        return cls(data=data, **kwargs)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.props.data is None:
            self.props.data = ([], [], [])

    def get_data_tuple(self) -> tuple[list, list, list]:
        """Get the data as a picklable tuple."""
        return self.props.data

    def set_data_tuple(self, data: tuple[list, list, list]) -> None:
        """Set the data from a tuple."""
        self.props.data = data


class ItemFactory(Graphs.ItemFactory):
    """Item factory."""

    _constructors = {
        "data-item": DataItem.new_with_data,
        "generated-data-item": GeneratedDataItem.new,
        "equation-item": EquationItem.new,
        "text-item": TextItem.new,
    }

    def __init__(self):
        super().__init__()
        for item, callback in self._constructors.items():
            self.connect(item + "-request", self._on_request, callback)

    @staticmethod
    def new_from_dict(dictionary: dict) -> Graphs.Item:
        """Instanciate item from dict."""
        match dictionary["type"]:
            case "DataItem":
                dictionary.pop("type")
                dictionary["data"] = Graphs.DataHolder.new(*dictionary["data"])
                return DataItem(**dictionary)
            case "GeneratedDataItem":
                dictionary.pop("type")
                dictionary["data"] = Graphs.DataHolder.new(*dictionary["data"])
                return GeneratedDataItem(**dictionary)
            case "EquationItem":
                cls = EquationItem
            case "TextItem":
                cls = TextItem
            case "FillItem":
                cls = FillItem
            case _:
                raise ValueError(f"could not find type {dictionary['type']}")
        dictionary.pop("type")
        return cls(**dictionary)

    @staticmethod
    def _on_request(self, data: Graphs.Data, *args) -> Graphs.Item:
        *args, callback = args
        return callback(data.get_selected_style_params(), *args)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Style editor."""
import asyncio
from gettext import gettext as _

from gi.repository import Gio, Graphs

from graphs.canvas import Canvas
from graphs.item import DataItem
from graphs.style_editor.editor_box import StyleEditorBox

import numpy

_PREVIEW_XDATA1 = numpy.linspace(0, 10, 10)
_PREVIEW_YDATA1 = numpy.linspace(0, numpy.power(numpy.e, 10), 10)
_PREVIEW_XERR1 = numpy.linspace(0.1, 0.5, 10)
_PREVIEW_YERR1 = numpy.linspace(500, 2500, 10)
_PREVIEW_XDATA2 = numpy.linspace(0, 10, 60)
_PREVIEW_YDATA2 = numpy.power(numpy.e, _PREVIEW_XDATA2)
CSS_TEMPLATE = """
.canvas-view#{name} {{
    background-color: {background_color};
    color: {color};
}}
"""


class PythonStyleEditor(Graphs.StyleEditor):
    """Graphs Style Editor Window."""

    __gtype_name__ = "GraphsPythonStyleEditor"

    def __init__(self):
        super().__init__()

        style_editor = StyleEditorBox(self)
        style_editor.connect("params-changed", self._on_params_changed)
        self.set_editor_box(style_editor)
        self._test_items = Gio.ListStore()
        self._initialize_test_items()
        self.connect("load_request", self._on_load_request)
        self.connect("save_request", self._on_save_request)

        self._background_task = asyncio.create_task(
            self._reload_canvas(style_editor),
        )

    def _initialize_test_items(self):
        """Initialize example test items with predefined preview data."""
        preview_data = [
            (_PREVIEW_XDATA1, _PREVIEW_YDATA1, _PREVIEW_XERR1, _PREVIEW_YERR1),
            (_PREVIEW_XDATA2, _PREVIEW_YDATA2, None, None),
        ]
        style_manager = Graphs.StyleManager.get_instance()
        test_style = style_manager.get_system_style_params()
        for xdata, ydata, xerr, yerr in preview_data:
            self._test_items.append(
                DataItem.new(
                    test_style,
                    xdata=xdata,
                    ydata=ydata,
                    xerr=xerr,
                    yerr=yerr,
                    name=_("Example Item"),
                    color="#000000",
                    errcolor="#000000",
                ),
            )

    def _on_params_changed(self, style_editor, changes_unsaved=True):
        self._background_task.cancel()
        self._background_task = asyncio.create_task(
            self._reload_canvas(style_editor, changes_unsaved, 0.5),
        )

    async def _reload_canvas(
        self,
        style_editor: StyleEditorBox,
        changes_unsaved: bool = False,
        timeout: int = 0,
    ) -> None:
        await asyncio.sleep(timeout)
        if style_editor.params is None:
            style_manager = Graphs.StyleManager.get_instance()
            params, graphs_params = style_manager.get_system_style_params()
        else:
            params = style_editor.params
            graphs_params = style_editor.graphs_params
            self.set_stylename(style_editor.graphs_params["name"])

        color_cycle = params["axes.prop_cycle"].by_key()["color"]
        errbar_cycle = graphs_params["errorbar.color_cycle"].by_key()["color"]
        for index, item in enumerate(self._test_items):
            item.set_color(color_cycle[index % len(color_cycle)])
            item.set_errcolor(errbar_cycle[index % len(errbar_cycle)])
            item.override((params, graphs_params))

        all_params = params, graphs_params
        canvas = Canvas(all_params, self._test_items, False)
        canvas.figure.props.title = _("Title")
        canvas.figure.props.bottom_label = _("X Label")
        canvas.figure.props.left_label = _("Y Label")
        self.set_canvas(canvas)

        # Set headerbar color
        css = CSS_TEMPLATE.format(
            name=self.props.content_view.get_name(),
            background_color=params["figure.facecolor"],
            color=params["text.color"],
        )
        self.props.css_provider.load_from_string(css)

        if changes_unsaved:
            self.set_unsaved(True)

    @staticmethod
    def _on_load_request(self, file: Gio.File) -> None:
        """Load a style."""
        style_editor = self.get_editor_box()
        name = style_editor.load_style(file)
        self.set_title(name)
        self._background_task = asyncio.create_task(
            self._reload_canvas(style_editor, False, 0),
        )

    @staticmethod
    def _on_save_request(self, file: Gio.File) -> None:
        """Save current style."""
        self.get_editor_box().save_style(file)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for parsing columns files."""
from gettext import pgettext as C_

from gi.repository import GLib, Graphs

from graphs.file_import.parsers import Parser
from graphs.misc import ParseError


class ColumnsParser(Parser):
    """Columns parser."""

    def __init__(self):
        super().__init__(
            "columns",
            C_("import-mode", "Columns"),
            None,
            None,
        )

    @staticmethod
    def parse(
        items: Graphs.ItemList,
        settings: Graphs.ImportSettings,
        data: Graphs.Data,
    ) -> None:
        """Import data from columns file."""
        parser = Graphs.ColumnsParser.new(settings)

        try:
            parser.parse()
            parser.add_items(data, items)
        except GLib.Error as e:
            raise ParseError(e.message) from e

    @staticmethod
    def init_settings_widgets(settings, box) -> None:
        """Append columns specific settings."""
        box.append(Graphs.ColumnsBox.new(settings))



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for style utilities."""
import io
import os

from gi.repository import Adw, GLib, Gdk, Gio, Graphs, Gtk

from graphs import style_io

from matplotlib import RcParams

CSS_TEMPLATE = """
.system-canvas-view {{
    background-color: {background_color};
    color: {color};
}}
"""


def _is_style_bright(params: RcParams):
    return Graphs.tools_get_luminance_from_hex(params["axes.facecolor"]) < 0.4


def _generate_preview(params: tuple[RcParams, dict]) -> Gdk.Texture:
    buffer = io.BytesIO()
    style_io.create_preview(buffer, params, "png", 31)
    return Gdk.Texture.new_from_bytes(GLib.Bytes.new(buffer.getvalue()))


def _params_for_bundled_style(name: str) -> tuple[RcParams, dict]:
    filename = Graphs.filename_from_stylename(name)
    uri = "resource:///se/sjoerd/Graphs/styles/" + filename
    return style_io.parse(Gio.File.new_for_uri(uri))


class StyleManager(Graphs.StyleManager):
    """
    Main Style Manager.

    Keeps track of all files in the style dir and represents them in
    the `selection_model` property.
    """

    __gtype_name__ = "GraphsPythonStyleManager"

    def __init__(self):
        # Check for Ubuntu
        gtk_theme = Gtk.Settings.get_default().get_property("gtk-theme-name")
        system_style_name = "Yaru" \
            if "SNAP" in os.environ \
            and gtk_theme.lower().startswith("yaru") \
            else "Adwaita"
        super().__init__()
        self.connect("style-request", self._on_style_request)
        self.connect("create-style-request", self._on_create_style_request)
        Adw.StyleManager.get_default().connect("notify", self._on_system_style)

        self._system_style_light_params = \
            _params_for_bundled_style(system_style_name)
        self._system_style_dark_params = \
            _params_for_bundled_style(system_style_name + " Dark")

        self._on_system_style()
        self.setup(system_style_name.lower())

    @staticmethod
    def _on_style_request(self, file: Gio.File) -> Graphs.Style:
        try:
            params = style_io.parse(
                file,
                self.get_system_style_params(),
            )
            style_params, graphs_params = params
            name = graphs_params["name"]
            preview = _generate_preview(params)
            light = _is_style_bright(style_params)
        except style_io.StyleParseError:
            name = ""
            preview = None
            light = False
        return Graphs.Style(
            name=name,
            file=file,
            mutable=True,
            preview=preview,
            light=light,
        )

    def get_system_style_params(self) -> tuple[RcParams, dict]:
        """Get the system style properties."""
        if Adw.StyleManager.get_default().get_dark():
            return self._system_style_dark_params
        else:
            return self._system_style_light_params

    def _on_system_style(self, *_args) -> None:
        params = self.get_system_style_params()
        css = CSS_TEMPLATE.format(
            background_color=params[0]["figure.facecolor"],
            color=params[0]["text.color"],
        )
        self.props.css_provider.load_from_string(css)

    @staticmethod
    def _on_create_style_request(
        self,
        template: Graphs.Style,
        destination: Gio.File,
        new_name: str,
    ) -> None:
        """Copy a style."""
        style_params, graphs_params = style_io.parse(
            template.get_file(),
            self.get_system_style_params(),
        )
        graphs_params["name"] = new_name
        style_io.write(destination, style_params, graphs_params)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Miscallaneous constants."""
import sympy


class ParseError(Exception):
    """Custom Error for parsing files."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class InvalidEquationError(Exception):
    """Custom Error for invalid equation."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


X = sympy.Symbol("x")

LINESTYLES = ["none", "solid", "dotted", "dashed", "dashdot"]
MARKERSTYLES = [
    "none",
    ".",
    ",",
    "o",
    "v",
    "^",
    "<",
    ">",
    "8",
    "s",
    "p",
    "*",
    "h",
    "H",
    "+",
    "x",
    "D",
    "d",
    "|",
    "_",
    "P",
    "X",
]

DIRECTIONS = ["bottom", "top", "left", "right"]

LIMITS = [
    f"{prefix}-{direction}" for direction in DIRECTIONS
    for prefix in ("min", "max")
]



# SPDX-License-Identifier: GPL-3.0-or-later
"""Various utility functions."""
from gi.repository import GLib, Graphs

import numexpr

import numpy


def bytes_to_ndarray(b: GLib.Bytes) -> numpy.ndarray:
    """Get a readonly ndarray referencing the original data."""
    if b is None:
        return None
    return numpy.frombuffer(b.get_data(), dtype=numpy.float64)


def create_equidistant_xdata(
    limits: tuple,
    scale: Graphs.Scale = Graphs.Scale.LINEAR,
    steps: int = 5000,
) -> numpy.ndarray:
    """Generate evenly-spaced x-values on the given scale."""
    x_start, x_stop = limits
    match scale:
        case Graphs.Scale.LINEAR | Graphs.Scale.RADIANS:
            xdata = numpy.linspace(x_start, x_stop, steps)
        case Graphs.Scale.LOG:
            x_start = max(x_start, 1e-300)
            x_stop = x_stop if numpy.isfinite(x_stop) else 1e300
            xdata = numpy.logspace(
                numpy.log10(x_start),
                numpy.log10(x_stop),
                steps,
            )
        case Graphs.Scale.LOG2:
            x_start = max(x_start, 1e-300)
            x_stop = x_stop if numpy.isfinite(x_stop) else 1e300
            xdata = numpy.logspace(
                numpy.log2(x_start),
                numpy.log2(x_stop),
                steps,
                base=2,
            )
        case Graphs.Scale.SQUAREROOT:
            x_start = max(x_start, 1e-300)
            xdata = numpy.linspace(
                numpy.sqrt(x_start),
                numpy.sqrt(x_stop),
                steps,
            )
            xdata = xdata**2
        case Graphs.Scale.INVERSE:
            xdata = (1 / numpy.linspace(1 / x_start, 1 / x_stop, steps))
        case _:
            raise ValueError
    return xdata


def equation_to_data(
    equation: str,
    limits: tuple,
    steps: int = 5000,
    scale: Graphs.Scale = Graphs.Scale.LINEAR,
) -> tuple:
    """Convert an equation into data over a specified range of x-values."""
    xdata = create_equidistant_xdata(limits, scale, steps)
    try:
        ydata = numexpr.evaluate(equation, local_dict={"x": xdata})
        if ydata.ndim == 0:
            ydata = numpy.full(steps, ydata)

        # Remove invalid values
        mask = numpy.isfinite(ydata)
        xdata = xdata[mask]
        ydata = ydata[mask]
    except (KeyError, SyntaxError, ValueError, TypeError):
        return None, None
    return xdata, ydata



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Module for parsing and writing styles.

This module is intended to be used at build time and thus must not depend on
other graphs modules.
"""
import contextlib
import logging
import typing
from gettext import gettext as _

from cycler import cycler

from gi.repository import Gio

from matplotlib import RcParams, cbook, rc_context
from matplotlib.figure import Figure
from matplotlib.font_manager import font_scalings, weight_dict
from matplotlib.style.core import STYLE_BLACKLIST

import numpy

STYLE_IGNORELIST = [
    "savefig.dpi",
    "savefig.facecolor",
    "savefig.edgecolor",
    "savefig.format",
    "savefix.bbox",
    "savefig.pad_inches",
    "savefig.transparent",
    "savefig.orientation",
]
FONT_SIZE_KEYS = [
    "font.size",
    "axes.labelsize",
    "xtick.labelsize",
    "ytick.labelsize",
    "legend.fontsize",
    "figure.labelsize",
    "figure.titlesize",
    "axes.titlesize",
]

STYLE_CUSTOM_PARAMS = [
    "name",
    "ticklabels",
    "errorbar.capsize",
    "errorbar.capthick",
    "errorbar.linewidth",
    "errorbar.barsabove",
    "errorbar.color_cycle",
]


class StyleParseError(Exception):
    """Custom Error for when a style cannot be parsed."""


def parse(
    file: Gio.File,
    validate: tuple[RcParams, dict] = None,
) -> (RcParams, dict):
    """
    Parse a style to RcParams.

    This is an improved version of matplotlibs '_rc_params_in_file()' function.
    It is also modified to work with GFile instead of the python builtin
    functions.
    """

    def _apply_defaults(
        target_dict: dict | RcParams,
        defaults: dict,
        filename,
    ) -> None:
        for key, value in defaults.items():
            if key not in target_dict:
                msg = (
                    "Parameter {key} not found in {filename}, using"
                    " default value: {value}"
                )
                msg = msg.format(key=key, filename=filename, value=value)
                logging.debug(msg)
                target_dict[key] = value

    style = RcParams()
    graphs_params = {"name": None}
    filename = file.get_basename()
    try:
        stream = Gio.DataInputStream.new(file.read(None))
        for line_number, line in enumerate(stream, 1):
            if line[:9] == "#~graphs ":
                graphs_param = True
                line = line[9:]
            else:
                graphs_param = False
            # legacy support for names at second line
            if line_number == 2 and graphs_params["name"] is None \
                    and line[:2] == "# ":
                graphs_params["name"] = line[2:]
            line = cbook._strip_comment(line)
            if not line:
                continue
            try:
                key, value = line.split(":", 1)
            except ValueError:
                msg = _("Missing colon in file {file}, line {line}")
                logging.warning(msg.format(file=filename, line=line_number))
                continue
            key = key.strip()
            value = value.strip()
            if value.startswith('"') and value.endswith('"'):
                value = value[1:-1]  # strip double quotes
            if key in STYLE_BLACKLIST:
                msg = _("Non-style related parameter {param} in file {file}")
                logging.warning(msg.format(param=key, file=filename))
            elif key in STYLE_IGNORELIST:
                msg = _("Ignoring parameter {param} in file {file}")
                logging.warning(msg.format(param=key, file=filename))
            elif key != "name" and \
                    key in (graphs_params if graphs_param else style):
                msg = _("Duplicate key in file {file}, on line {line}")
                logging.warning(msg.format(file=filename, line=line_number))
            else:
                if key in FONT_SIZE_KEYS \
                        and not value.replace(".", "", 1).isdigit():
                    try:
                        value = font_scalings[value]
                    except KeyError:
                        continue
                elif key == "font.weight" and not value.isdigit():
                    try:
                        value = weight_dict[value]
                    except KeyError:
                        continue
                try:
                    if graphs_param:
                        if key not in STYLE_CUSTOM_PARAMS:
                            msg = _(
                                "Bad value in file {file} on line {line},"
                                " custom parameter {key} is not supported",
                            )
                            logging.warning(
                                msg.format(
                                    file=filename,
                                    line=line_number,
                                    key=key,
                                ),
                            )
                            continue
                        # Convert boolean-strings to boolean:
                        bool_mapping = {"false": False, "true": True}
                        value = bool_mapping.get(value.lower(), value)
                        if key == "errorbar.color_cycle":
                            color = ["#" + c.strip() for c in value.split(",")]
                            value = cycler(color=color)
                        else:
                            with contextlib.suppress(ValueError):
                                value = float(value)
                        graphs_params[key] = value
                    else:
                        style[key] = value
                except (KeyError, ValueError):
                    msg = _("Bad value in file {file} on line {line}")
                    logging.warning(
                        msg.format(file=filename, line=line_number),
                    )
    except UnicodeDecodeError as error:
        msg = _("Could not parse {filename}").format(filename=filename)
        raise StyleParseError(msg) from error
    finally:
        stream.close()
    if validate is not None:
        style_defaults, graph_defaults = validate
        _apply_defaults(style, style_defaults, filename)
        _apply_defaults(graphs_params, graph_defaults, filename)

    if graphs_params["name"] is None:
        msg = _("File {file}, does not contain name tag")
        logging.warning(msg.format(file=filename))
        graphs_params["name"] = filename
    return style, graphs_params


WRITE_IGNORELIST = STYLE_IGNORELIST + [
    "lines.dashdot_pattern",
    "lines.dashed_pattern",
    "lines.dotted_pattern",
    "lines.dash_capstyle",
    "lines.dash_joinstyle",
    "lines.solid_capstyle",
    "lines.solid_joinstyle",
]


def write(file: Gio.File, style: RcParams, graphs_params: dict) -> None:
    """Write a style to a file."""
    stream = Gio.DataOutputStream.new(file.replace(None, False, 0, None))
    stream.put_string("# Generated via Graphs\n")
    for key, value in graphs_params.items():
        if key == "errorbar.color_cycle":
            value = ", ".join(c.lstrip("#") for c in value.by_key()["color"])
        stream.put_string(f"#~graphs {key}: {value}\n")
    for key, value in style.items():
        if key not in STYLE_BLACKLIST and key not in WRITE_IGNORELIST:
            value = str(value).replace("#", "")
            if key != "axes.prop_cycle":
                value = value.replace("[", "").replace("]", "")
                value = value.replace("'", "").replace("'", "")
                value = value.replace('"', "").replace('"', "")
            stream.put_string(f"{key}: {value}\n")
    stream.close()


_PREVIEW_XDATA = numpy.linspace(0, 10, 30)
_PREVIEW_YDATA1 = numpy.sin(_PREVIEW_XDATA)
_PREVIEW_YDATA2 = numpy.cos(_PREVIEW_XDATA)


def create_preview(
    file: typing.IO,
    params: tuple[RcParams, dict],
    file_format: str = "svg",
    dpi: int = 100,
) -> None:
    """Create preview of params and write it to file."""
    style_params, graphs_params = params
    with rc_context(style_params):
        # set render size in inch
        figure = Figure(figsize=(5, 3))
        axis = figure.add_subplot()
        axis.spines.bottom.set_visible(True)
        axis.spines.left.set_visible(True)
        if style_params["axes.spines.top"] and graphs_params["ticklabels"]:
            tick_params = {
                "labelleft": style_params["ytick.left"],
                "labelright": style_params["ytick.right"],
                "labeltop": style_params["xtick.top"],
                "labelbottom": style_params["xtick.bottom"],
            }
            axis.tick_params(which="both", **tick_params)
        else:
            axis.tick_params(which="both", top=False, right=False)
        axis.plot(_PREVIEW_XDATA, _PREVIEW_YDATA1)
        axis.plot(_PREVIEW_XDATA, _PREVIEW_YDATA2)
        axis.set_xlabel(_("X Label"))
        axis.set_xlabel(_("Y Label"))
        figure.savefig(file, format=file_format, dpi=dpi)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Data management module."""
import copy
import logging
from collections import OrderedDict
from collections.abc import Iterator
from gettext import gettext as _
from operator import itemgetter

from gi.repository import Gio, Graphs, Gtk

from graphs import misc, project, style_io, utilities
from graphs.item import ItemFactory

from matplotlib import RcParams

import numexpr

import numpy

import sympy
from sympy.calculus.singularities import singularities

_FIGURE_SETTINGS_HISTORY_IGNORELIST = misc.LIMITS + [
    "min-selected",
    "max-selected",
]

LOG_SCALES = {1, 2}
NONZERO_SCALES = {1, 2, 4}


class Data(Graphs.Data):
    """Class for managing data."""

    __gtype_name__ = "GraphsPythonData"

    def __init__(self):
        self._selected_style_params = None, {}
        super().__init__()
        self.connect("load-request", self._on_load_request)
        self.connect("position-changed", self._on_position_changed)
        self.connect("item-changed", self._on_item_changed)
        self.connect("item-added", self._on_item_added)
        self.connect("item-removed", self._on_item_removed)
        self.connect(
            "figure-settings-changed",
            self._on_figure_settings_changed,
        )
        self.connect(
            "add-history-state-request",
            self._on_add_history_state_request,
        )

    def __len__(self) -> int:
        """Magic alias for `get_n_items()`."""
        return self.get_n_items()

    def __iter__(self) -> Iterator[Graphs.Item]:
        """Iterate over items."""
        for i in range(self.get_n_items()):
            yield self.get_item(i)

    def __getitem__(self, pos: int):
        """Magic alias for retrieving items."""
        return self.get_item(pos)

    def get_old_selected_style_params(self) -> tuple[RcParams, dict]:
        """Get the old selected style properties."""
        return self._old_style_params

    def get_selected_style_params(self) -> tuple[RcParams, dict]:
        """Get the selected style properties."""
        return self._selected_style_params

    def _update_selected_style(self) -> None:
        figure_settings = self.props.figure_settings
        style_manager = Graphs.StyleManager.get_instance()
        error_msg = None
        if figure_settings.get_use_custom_style():
            stylename = figure_settings.get_custom_style()
            for style in self.props.style_selection_model.get_model():
                if stylename == style.get_name():
                    try:
                        validate = None
                        if style.get_mutable():
                            validate = style_manager.get_system_style_params()
                        self._old_style_params = self._selected_style_params

                        style_params, graphs_params = style_io.parse(
                            style.get_file(),
                            validate,
                        )
                        self._selected_style_params = \
                            style_params, graphs_params

                        self.set_color_cycle(
                            style_params["axes.prop_cycle"].by_key()["color"],
                        )
                        self.set_errbar_color_cycle(
                            graphs_params["errorbar.color_cycle"].by_key()
                            ["color"],
                        )
                        return
                    except (ValueError, SyntaxError, AttributeError):
                        error_msg = _(
                            "Could not parse style {stylename}, loading "
                            "system preferred style",
                        ).format(stylename=stylename)
                    break
            error_msg = _(
                "Style {stylename} does not exist, "
                "loading system preferred style",
            ).format(stylename=stylename)

        if error_msg is not None:
            figure_settings.set_use_custom_style(False)
            logging.warning(error_msg)

        self._old_style_params = self._selected_style_params
        self._selected_style_params = style_manager.get_system_style_params()
        color_cycle = self._selected_style_params[0]["axes.prop_cycle"]
        self.set_color_cycle(color_cycle.by_key()["color"])
        graphs_params = self._selected_style_params[1]
        self.set_errbar_color_cycle(
            graphs_params["errorbar.color_cycle"].by_key()["color"],
        )

    def _init_history_states(self) -> None:
        limits = self.props.figure_settings.get_limits().values()
        self._history_states = [([], limits)]
        self._history_pos = -1
        self._set_data_copy()

    @staticmethod
    def _on_position_changed(self, index1: int, index2: int) -> None:
        """Change item position of index2 to that of index1."""
        self._current_batch.append((
            Graphs.ChangeType.ITEMS_SWAPPED,
            (index2, index1),
        ))

    @staticmethod
    def _on_item_added(self, item: Graphs.Item) -> None:
        self._current_batch.append((
            Graphs.ChangeType.ITEM_ADDED,
            item.to_dict(),
        ))

    @staticmethod
    def _on_item_removed(self, item: Graphs.Item, index: int) -> None:
        self._current_batch.append((
            Graphs.ChangeType.ITEM_REMOVED,
            (index, item.to_dict()),
        ))

    @staticmethod
    def _on_item_changed(self, item: Graphs.Item, prop: str) -> None:
        index = self.index(item)
        value = item.get_property(prop) if prop != "data" \
            else item.get_data_tuple()
        self._current_batch.append((
            Graphs.ChangeType.ITEM_PROPERTY_CHANGED,
            (
                index,
                prop,
                copy.deepcopy(self._data_copy[index][prop]),
                value,
            ),
        ))

    @staticmethod
    def _on_figure_settings_changed(self, prop: str) -> None:
        if prop in _FIGURE_SETTINGS_HISTORY_IGNORELIST:
            return
        self._current_batch.append((
            Graphs.ChangeType.FIGURE_SETTINGS_CHANGED,
            (
                prop,
                copy.deepcopy(self._figure_settings_copy[prop]),
                copy.deepcopy(self.props.figure_settings.get_property(prop)),
            ),
        ))

    def _set_data_copy(self) -> None:
        """Set a deep copy for the data."""
        self._current_batch: list = []
        self._data_copy = copy.deepcopy([item.to_dict() for item in self])
        self._figure_settings_copy = copy.deepcopy({
            prop.replace("_", "-"):
            self.props.figure_settings.get_property(prop)
            for prop in dir(self.props.figure_settings.props)
        })

    def _collapse_current_batch(self) -> None:
        """
        Collapse transitive changes within the current history batch.

        This method reduces redundant "change" entries in the current history
        batch:
        - Multiple consecutive item and figure settings changes to the same
          property are collapsed into a single change.
        - If a collapsed change would be meaningless (the original old value
          is equal to the most recent new value), that change is dropped.
        - If the batch contains structural changes like items added or removed,
          the method aborts and leaves the batch unchanged.
        """
        if not self._current_batch:
            # Nothing to collapse
            return

        collapsed = OrderedDict()

        for change_type, data in self._current_batch:
            match change_type:
                case Graphs.ChangeType.ITEM_PROPERTY_CHANGED:
                    index, prop, _old_value, new_value = data
                    key = (change_type, index, prop)

                    if key not in collapsed:
                        collapsed[key] = (change_type, data)
                    else:
                        first_old = collapsed[key][1][2]
                        if first_old == new_value:
                            collapsed.pop(key)
                        else:
                            collapsed[key] = (
                                Graphs.ChangeType.ITEM_PROPERTY_CHANGED,
                                (index, prop, first_old, new_value),
                            )

                case Graphs.ChangeType.FIGURE_SETTINGS_CHANGED:
                    prop, _old_value, new_value = data
                    key = (change_type, prop)

                    if key not in collapsed:
                        collapsed[key] = (change_type, data)
                    else:
                        first_old = collapsed[key][1][1]
                        if first_old == new_value:
                            collapsed.pop(key)
                        else:
                            collapsed[key] = (
                                Graphs.ChangeType.FIGURE_SETTINGS_CHANGED,
                                (prop, first_old, new_value),
                            )

                case _:
                    # On any other change such as items added or removed we
                    # abort collapsing
                    return

        self._current_batch = list(collapsed.values())

    @staticmethod
    def _on_add_history_state_request(self) -> bool:
        """Add a state to the clipboard."""
        self._collapse_current_batch()
        if not self._current_batch:
            # Nothing to add
            return False
        if self._history_pos != -1:
            self._history_states = self._history_states[:self._history_pos + 1]
        self._history_pos = -1
        limits = self.get_figure_settings().get_limits().values()
        self._history_states.append((self._current_batch, limits))
        # Keep history states length limited to 100 spots
        if len(self._history_states) > 101:
            self._history_states = self._history_states[1:]
        self._set_data_copy()
        return True

    def _undo(self) -> None:
        """Undo the latest change that was added to the clipboard."""
        if not self.props.can_undo:
            return
        batch = self._history_states[self._history_pos][0]
        self._history_pos -= 1
        selected = Gtk.Bitset.new_empty()
        mask = Gtk.Bitset.new_empty()
        for change_type, change in reversed(batch):
            match change_type:
                case Graphs.ChangeType.ITEM_PROPERTY_CHANGED:
                    index, prop, value = itemgetter(0, 1, 2)(change)
                    if prop == "selected":
                        mask.add(index)
                        if value:
                            selected.add(index)
                    elif prop == "data":
                        self[index].set_data_tuple(value)
                    else:
                        self[index].set_property(prop, value)
                case Graphs.ChangeType.ITEM_ADDED:
                    self._remove_item(self.get_n_items() - 1)
                case Graphs.ChangeType.ITEM_REMOVED:
                    self._insert_item(
                        ItemFactory.new_from_dict(copy.deepcopy(change[1])),
                        change[0],
                    )
                case Graphs.ChangeType.ITEMS_SWAPPED:
                    self.change_position(change[0], change[1])
                case Graphs.ChangeType.FIGURE_SETTINGS_CHANGED:
                    self.props.figure_settings.set_property(
                        change[0],
                        change[1],
                    )
        self.set_selection(selected, mask)
        limits = Graphs.Limits.new(self._history_states[self._history_pos][1])
        self.get_figure_settings().set_limits(limits)
        self.props.can_redo = True
        self.props.can_undo = \
            abs(self._history_pos) < len(self._history_states)
        self._set_data_copy()

    def _redo(self) -> None:
        """Redo the latest change that was added to the clipboard."""
        if not self.props.can_redo:
            return
        self._history_pos += 1
        state = self._history_states[self._history_pos]
        selected = Gtk.Bitset.new_empty()
        mask = Gtk.Bitset.new_empty()
        for change_type, change in state[0]:
            match change_type:
                case Graphs.ChangeType.ITEM_PROPERTY_CHANGED:
                    index, prop, value = itemgetter(0, 1, 3)(change)
                    if prop == "selected":
                        mask.add(index)
                        if value:
                            selected.add(index)
                    elif prop == "data":
                        self[index].set_data_tuple(value)
                    else:
                        self[index].set_property(prop, value)
                case Graphs.ChangeType.ITEM_ADDED:
                    change = copy.deepcopy(change)
                    self._add_item(ItemFactory.new_from_dict(change))
                case Graphs.ChangeType.ITEM_REMOVED:
                    self._remove_item(change[0])
                case Graphs.ChangeType.ITEMS_SWAPPED:
                    self.change_position(change[1], change[0])
                case Graphs.ChangeType.FIGURE_SETTINGS_CHANGED:
                    self.props.figure_settings.set_property(
                        change[0],
                        change[2],
                    )
        self.set_selection(selected, mask)
        self.get_figure_settings().set_limits(Graphs.Limits.new(state[1]))
        self.props.can_redo = self._history_pos < -1
        self.props.can_undo = True
        self._set_data_copy()

    def _optimize_limits(self) -> None:
        """Optimize the limits of the canvas to the data class."""
        figure_settings = self.get_figure_settings()

        axes = [[
            direction,
            False,
            figure_settings.get_property(f"min_{direction}"),
            figure_settings.get_property(f"max_{direction}"),
            figure_settings.get_property(f"{direction}_scale"),
            None,
        ] for direction in ("bottom", "left", "top", "right")]

        equation_items = []
        hide_unselected = figure_settings.get_hide_unselected()

        for item in self:
            if not isinstance(item, (Graphs.DataItem, Graphs.EquationItem)):
                continue

            if not item.get_selected() and hide_unselected:
                continue

            if isinstance(item, Graphs.EquationItem):
                equation_items.append(item)
                continue

            indices = (item.get_xposition() * 2, 1 + item.get_yposition() * 2)
            for index, xydata in zip(indices, item.get_xydata()):
                axis = axes[index]

                xydata = xydata[numpy.isfinite(xydata)]

                if xydata.size == 0:
                    continue

                if axis[4] in NONZERO_SCALES:
                    nonzero = xydata[xydata != 0]
                    min_value = nonzero.min() if nonzero.size else xydata.min()
                else:
                    min_value = xydata.min()
                max_value = xydata.max()

                if axis[1]:
                    axis[2] = min(axis[2], min_value)
                    axis[3] = max(axis[3], max_value)
                else:
                    axis[2] = min_value
                    axis[3] = max_value
                    axis[1] = True

        for item in equation_items:
            xindex = item.get_xposition() * 2
            xaxis = axes[xindex]
            yaxis = axes[1 + item.get_yposition() * 2]
            x_limits = [xaxis[2], xaxis[3]]
            yscale = yaxis[4]

            equation = item.get_preprocessed_equation()
            expr = sympy.sympify(equation)
            domain = sympy.Interval(*x_limits)
            has_singularities = singularities(expr, misc.X, domain)

            if xaxis[5] is None:
                xscale = xaxis[4]
                xaxis[5] = utilities.create_equidistant_xdata(x_limits, xscale)
            ydata = numexpr.evaluate(equation, local_dict={"x": xaxis[5]})
            ydata = ydata[numpy.isfinite(ydata)]

            if has_singularities:
                # Don't take negative values into account for log scaling
                if yscale in LOG_SCALES:
                    ydata = ydata[ydata > 0]

                y_min, y_max = ydata.min(), ydata.max()
                lower_bound = Graphs.get_value_at_fraction(
                    0.05,
                    y_min,
                    y_max,
                    yscale,
                )
                upper_bound = Graphs.get_value_at_fraction(
                    0.95,
                    y_min,
                    y_max,
                    yscale,
                )
                ydata = ydata.clip(lower_bound, upper_bound)

            if ydata.size == 0:
                continue

            if yscale in NONZERO_SCALES:
                nonzero = ydata[ydata != 0]
                min_value = nonzero.min() if nonzero.size else ydata.min()
            else:
                min_value = ydata.min()
            max_value = ydata.max()

            if yaxis[1]:
                yaxis[2] = min(yaxis[2], min_value)
                yaxis[3] = max(yaxis[3], max_value)
            else:
                yaxis[2] = min_value
                yaxis[3] = max_value
                yaxis[1] = True

        for count, (direction, used, min_all, max_all, scale, _x) in \
                enumerate(axes):
            if not used:
                continue

            # 0.05 padding on y-axis, 0.015 padding on x-axis
            padding_factor = 0.05 if count % 2 else 0.015

            if scale not in LOG_SCALES:  # For non-logarithmic scales
                span = max_all - min_all
                max_all += padding_factor * span

                # For inverse scale, calculate padding using a factor
                if scale == 4:
                    min_all *= 0.99
                else:
                    min_all -= padding_factor * span
            else:  # Use different scaling type for logarithmic scale
                log_min = numpy.log10(min_all) if min_all > 0 else 0
                log_max = numpy.log10(max_all) if max_all > 0 else 0
                log_span = log_max - log_min

                log_min -= padding_factor * log_span
                log_max += padding_factor * log_span
                min_all = 10**log_min
                max_all = 10**log_max
            figure_settings.set_property(f"min_{direction}", min_all)
            figure_settings.set_property(f"max_{direction}", max_all)

    def get_project_dict(self) -> dict:
        """Convert data to dict."""
        figure_settings = self.get_figure_settings()
        view_pos, view_states = self.get_view_history()
        return {
            "version": self.get_version(),
            "data": [item.to_dict() for item in self],
            "figure-settings": {
                key.replace("_", "-"): figure_settings.get_property(key)
                for key in dir(figure_settings.props)
            },
            "history-states": self._history_states,
            "history-position": self._history_pos,
            "view-history-states": [lims.values() for lims in view_states],
            "view-history-position": view_pos,
        }

    def load_from_project_dict(self, project_dict: dict) -> None:
        """Load data from dict."""
        # Load data
        self.set_figure_settings(
            Graphs.FigureSettings(
                **{
                    key.replace("-", "_"): value
                    for (key, value) in project_dict["figure-settings"].items()
                },
            ),
        )
        items = list(map(ItemFactory.new_from_dict, project_dict["data"]))
        self.set_items(items)

        # Set clipboard
        self._set_data_copy()
        self._history_states = project_dict["history-states"]
        self._history_pos = project_dict["history-position"]
        view_states = project_dict["view-history-states"]
        limits = list(map(Graphs.Limits.new, view_states))
        self.set_view_history(project_dict["view-history-position"], limits)

        # Set clipboard/view buttons
        self.props.can_undo = \
            abs(self._history_pos) < len(self._history_states)
        self.props.can_redo = self._history_pos < -1

    def _save(self) -> None:
        project.save_project_dict(self.props.file, self.get_project_dict())

    @staticmethod
    def _on_load_request(
        self,
        file: Gio.File,
        parse_flags: Graphs.ProjectParseFlags,
    ) -> str:
        try:
            project_dict = project.read_project_file(file, parse_flags)
        except project.ProjectParseError as error:
            if error.log:
                logging.exception(error)
            return error.message
        current_data = self.get_project_dict()
        try:
            self.load_from_project_dict(project_dict)
        except Exception:
            self.load_from_project_dict(current_data)
            msg = _("Failed to load project")
            logging.exception(msg)
            return msg
        return ""



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for saving and loading projects."""
import copy
import logging
import re
from gettext import gettext as _
from operator import itemgetter

from gi.repository import Gio, Graphs

from graphs import file_io, migrate
from graphs.item import ItemFactory

CURRENT_PROJECT_VERSION = 2


class ProjectParseError(Exception):
    """Custom error for parsing projects."""

    def __init__(self, message: str, log: bool = True):
        super().__init__()
        self.message = message
        self.log = log


PROJECT_KEYS = [
    "version",
    "data",
    "figure-settings",
    "history-states",
    "history-position",
    "view-history-states",
    "view-history-position",
]


class ProjectMigrator:
    """
    Migrate project data to be compatible with the current version of Graphs.

    Migration is based on version to version basis, so migrating from v1 to v3
    would involve migrating from v1 to v2 first and then migrating from v2
    to v3.
    """

    beta_pattern = re.compile(r"^\d+\.\d+\.\d+-[0-9a-f]{8}$")

    def __init__(
        self,
        project_dict: dict,
        parse_flags: Graphs.ProjectParseFlags,
    ):
        # Verify all keys are present
        try:
            for key in PROJECT_KEYS:
                project_dict[key]
        except KeyError as e:
            raise ProjectParseError(_("Project file is missing data")) from e

        self._project_dict = project_dict
        self.parse_flags = parse_flags

        # check beta version
        beta_version = bool(self.beta_pattern.match(project_dict["version"]))
        check_beta = not parse_flags & Graphs.ProjectParseFlags.ALLOW_BETA
        if beta_version and check_beta:
            raise ProjectParseError("BETA_DISALLOWED", False)

    def migrate(self) -> dict:
        """Perform needed migrations."""
        try:
            project_version = int(self._project_dict["project-version"])
            assert project_version > 0
        except KeyError:
            project_version = 1

        if project_version > CURRENT_PROJECT_VERSION:
            raise ProjectParseError(
                _("Project is from a newer incompatible version of Graphs"),
            )

        if CURRENT_PROJECT_VERSION == project_version:
            return self._project_dict

        # Migrate a project one version at a time
        current_version = project_version
        while current_version < CURRENT_PROJECT_VERSION:
            current_version += 1
            getattr(self, f"_migrate_v{current_version}")()
        return self._project_dict

    def _migrate_v2(self):
        logging.debug("migrating project v1 to v2")
        # Figure settings entries are now properly stored in hyphon-case
        self._project_dict["figure-settings"] = {
            key.replace("_", "-"): value
            for (key, value) in self._project_dict["figure-settings"].items()
        }

        self._migrate_inserted_scale(2)  # log2 scale added

        # Handle items no longer making use of uuid as well as xdata and ydata
        # being stored in a tuple as data. Also truncate the Graphs prefix
        # before the typename
        def _item_dict(item):
            return_item = {
                key: value
                for (key, value) in item.items()
                if key not in ("uuid", "xdata", "ydata")
            }
            return_item["data"] = (item["xdata"], item["ydata"], None, None)
            return_item["type"] = return_item["type"][6:]
            return return_item

        item_positions = []
        data = []
        for item in self._project_dict["data"]:
            item_positions.append(item["uuid"])
            data.append(_item_dict(item))
        self._project_dict["data"] = data
        history_states = self._project_dict["history-states"]
        n_states = len(history_states)
        history_pos = self._project_dict["history-position"]
        while history_pos < 1:
            for (change_type, change) in history_states[history_pos][0]:
                match change_type:
                    case Graphs.ChangeType.ITEM_ADDED:
                        item_positions.append(change["uuid"])
                    case Graphs.ChangeType.ITEM_REMOVED:
                        item_positions.remove(change[1]["uuid"])
                    case Graphs.ChangeType.ITEMS_SWAPPED:
                        uuid = item_positions.pop(change[0])
                        item_positions.insert(change[1], uuid)
            history_pos += 1
        data_changes = {}
        for state_index, state in enumerate(reversed(history_states)):
            state_index = n_states - state_index - 1
            new_state = []
            for change_type, change in reversed(state[0]):
                match change_type:
                    case Graphs.ChangeType.ITEM_ADDED:
                        item_positions.remove(change["uuid"])
                        change = _item_dict(change)
                    case Graphs.ChangeType.ITEM_REMOVED:
                        item_positions.insert(change[0], change[1]["uuid"])
                        change = [change[0], _item_dict(change[1])]
                    case Graphs.ChangeType.ITEMS_SWAPPED:
                        uuid = item_positions.pop(change[1])
                        item_positions.insert(change[0], uuid)
                    case Graphs.ChangeType.ITEM_PROPERTY_CHANGED:
                        change[0] = item_positions.index(change[0])
                        if change[1] in ("xdata", "ydata"):
                            # Consolidate two change entries into a single one
                            try:
                                data_change = data_changes[state_index]
                                if change[1] == "xdata":
                                    data_change[0] = change[2:]
                                else:
                                    data_change[1] = change[2:]

                                change = [
                                    Graphs.ChangeType.ITEM_PROPERTY_CHANGED,
                                    "data",
                                    (
                                        data_change[0][0],
                                        data_change[1][0],
                                        None,
                                        None,
                                    ),
                                    (
                                        data_change[0][1],
                                        data_change[1][1],
                                        None,
                                        None,
                                    ),
                                ]
                            except KeyError:
                                if change[1] == "xdata":
                                    xdata = change[2:]
                                    ydata = None
                                else:
                                    xdata = None
                                    ydata = change[2:]
                                data_changes[state_index] = [xdata, ydata]
                                continue
                new_state.insert(0, [change_type, change])
            history_states[state_index][0] = new_state
        self._project_dict["history-states"] = history_states

    def _migrate_inserted_scale(self, scale_index: int) -> None:
        """Handle a new scale being inserted at scale_index."""
        figure_settings = self._project_dict["figure-settings"]
        for prefix in ("left", "right", "top", "bottom"):
            axis = prefix + "-scale"
            if figure_settings[axis] >= scale_index:
                figure_settings[axis] = figure_settings[axis] + 1

        for state_index, history_state in enumerate(
            self._project_dict["history-states"],
        ):
            for change_index, changeset in enumerate(history_state[0]):
                change_type, change = changeset
                if change_type != 4:
                    continue
                if change[0][-6:] != "-scale":
                    continue
                for i, val in enumerate(change[1:], 1):
                    if val >= scale_index:
                        self._project_dict["history-states"][state_index][0][
                            change_index][1][i] = val + 1


class ProjectValidator:
    """Validate the project."""

    def __init__(
        self,
        project_dict: dict,
        parse_flags: Graphs.ProjectParseFlags,
    ):
        self.project_dict = copy.deepcopy(project_dict)
        self.parse_flags = parse_flags

    def validate(self):
        """Run through the history states."""
        # Validate Figure Settings
        self.figure_settings = Graphs.FigureSettings(
            **{
                key.replace("-", "_"): val
                for (key, val) in self.project_dict["figure-settings"].items()
            },
        )

        # Validate items
        data = self.project_dict["data"]
        self.items = [ItemFactory.new_from_dict(d) for d in data]

        # Validate view history
        view_history_states = self.project_dict["view-history-states"]
        for history_state in view_history_states:
            self.figure_settings.set_limits(Graphs.Limits.new(history_state))
        view_history_pos = int(self.project_dict["view-history-position"])
        assert view_history_pos < 0
        assert abs(view_history_pos) <= len(view_history_states)

        # Validate data history
        history_states = self.project_dict["history-states"]
        history_pos = int(self.project_dict["history-position"])
        assert history_pos < 0
        assert abs(history_pos) <= len(history_states)
        while history_pos < -1:
            for (change_type, change) in history_states[history_pos][0]:
                match change_type:
                    case Graphs.ChangeType.ITEM_PROPERTY_CHANGED:
                        index, prop, value = itemgetter(0, 1, 3)(change)
                        if prop == "data":
                            self.items[index].set_data_tuple(value)
                        else:
                            self.items[index].set_property(prop, value)
                    case Graphs.ChangeType.ITEM_ADDED:
                        data = copy.deepcopy(change)
                        self.items.append(ItemFactory.new_from_dict(data))
                    case Graphs.ChangeType.ITEM_REMOVED:
                        self.items.pop(change[0])
                    case Graphs.ChangeType.ITEMS_SWAPPED:
                        self.items.insert(change[1], self.items.pop(change[0]))
                    case Graphs.ChangeType.FIGURE_SETTINGS_CHANGED:
                        self.figure_settings.set_property(change[0], change[2])
            history_pos += 1
        for history_state in reversed(history_states):
            limits = history_state[1]
            self.figure_settings.set_limits(Graphs.Limits.new(limits))
            for change_type, change in reversed(history_state[0]):
                match change_type:
                    case Graphs.ChangeType.ITEM_PROPERTY_CHANGED:
                        index, prop, value = itemgetter(0, 1, 2)(change)
                        if prop == "data":
                            self.items[index].set_data_tuple(value)
                        else:
                            self.items[index].set_property(prop, value)
                    case Graphs.ChangeType.ITEM_ADDED:
                        self.items.pop()
                    case Graphs.ChangeType.ITEM_REMOVED:
                        data = copy.deepcopy(change[1])
                        item = ItemFactory.new_from_dict(data)
                        self.items.insert(change[0], item)
                    case Graphs.ChangeType.ITEMS_SWAPPED:
                        self.items.insert(change[0], self.items.pop(change[1]))
                    case Graphs.ChangeType.FIGURE_SETTINGS_CHANGED:
                        self.figure_settings.set_property(change[0], change[1])


def read_project_file(
    file: Gio.File,
    parse_flags: Graphs.ProjectParseFlags = Graphs.ProjectParseFlags.NONE,
) -> dict:
    """Read a project dict from file and account for migration."""
    try:
        project_dict = file_io.parse_json(file)
    except UnicodeDecodeError:
        if not parse_flags & Graphs.ProjectParseFlags.ALLOW_LEGACY_MIGRATION:
            raise ProjectParseError("LEGACY_MIGRATION_DISALLOWED", False)
        try:
            project_dict = migrate.migrate_project(file)
        except Exception as e:
            raise ProjectParseError(_("Failed to do legacy migration")) from e
    except Exception as e:
        raise ProjectParseError(_("Failed to parse project file")) from e
    try:
        project_dict = ProjectMigrator(project_dict, parse_flags).migrate()
    except ProjectParseError:
        raise
    except Exception as e:
        raise ProjectParseError(_("Failed to migrate project")) from e
    try:
        ProjectValidator(project_dict, parse_flags).validate()
    except ProjectParseError:
        raise
    except Exception as e:
        raise ProjectParseError(_("Failed to validate project")) from e
    return project_dict


def save_project_dict(file: Gio.File, project_dict: dict) -> None:
    """Save a project dict to a file."""
    project_dict["project-version"] = CURRENT_PROJECT_VERSION
    file_io.write_json(file, project_dict, pretty_print=False)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Scales Module."""
from matplotlib import scale, ticker, transforms

import numpy


class Log2Scale(scale.LogScale):
    """
    Logarithmic Base 2 Scale.

    Small utility for implementing logarithmic base 2 scaling easily using
    the scale name system.
    """

    name = "log2"

    def __init__(self, axis, *args, **kwargs):
        kwargs["base"] = 2
        super().__init__(axis, *args, **kwargs)


class RadiansScale(scale.LinearScale):
    """Radians Scale."""

    name = "radians"

    def set_default_locators_and_formatters(self, axis):
        """Handle locators and formatters."""
        super().set_default_locators_and_formatters(axis)
        axis.set_major_formatter(
            ticker.FuncFormatter(lambda x, _pos=None: f"{x / numpy.pi:.3g}π"),
        )
        axis.set_major_locator(RadianLocator())


class SquareRootScale(scale.ScaleBase):
    """Class for generating custom square root scale."""

    name = "squareroot"

    def set_default_locators_and_formatters(self, axis):
        """Handle locators and formatters."""
        axis.set_major_locator(CustomScaleLocator())
        axis.set_minor_locator(CustomScaleLocator(is_minor=True))
        axis.set_major_formatter(ticker.ScalarFormatter())
        axis.set_minor_formatter(ticker.NullFormatter())

    def limit_range_for_scale(self, vmin, vmax, _minpos):
        """Limit scale range."""
        return max(0, vmin), vmax

    class SquareRootTransform(transforms.Transform):
        """The transform to convert from linear to square root scale."""

        input_dims = 1  # Amount of input params in transform
        output_dims = 1  # Amount of output params in transform
        is_separable = True  # Seperable in X and Y dimension

        def transform_non_affine(self, a):
            """Transform data."""
            # Don't spam about invalid divide by zero errors
            with numpy.errstate(divide="ignore", invalid="ignore"):
                return numpy.array(a)**0.5

        def inverted(self):
            """Get the inverse transform."""
            return SquareRootScale.InvertedSquareRootTransform()

    class InvertedSquareRootTransform(transforms.Transform):
        """Inverse transform to convert from square root to linear scale."""

        input_dims = 1  # Amount of input params in transform
        output_dims = 1  # Amount of output params in transform
        is_separable = True  # Seperable in X and Y dimension

        def transform_non_affine(self, a):
            """Transform data."""
            # Don't spam about invalid divide by zero errors
            with numpy.errstate(divide="ignore", invalid="ignore"):
                return numpy.array(a)**2

        def inverted(self):
            """Get the inverse transform."""
            return SquareRootScale.SquareRootTransform()

    def get_transform(self):
        """Get the transform."""
        return self.SquareRootTransform()


class InverseScale(scale.ScaleBase):
    """Inverse scale."""

    name = "inverse"

    def set_default_locators_and_formatters(self, axis):
        """Handle locators and formatters."""
        axis.set_major_locator(CustomScaleLocator())
        axis.set_minor_locator(CustomScaleLocator(is_minor=True))
        axis.set_major_formatter(ticker.ScalarFormatter())
        axis.set_minor_formatter(ticker.NullFormatter())

    def limit_range_for_scale(self, vmin, vmax, minpos):
        """Limit scale range."""
        if not numpy.isfinite(minpos):
            minpos = 1e-300
        return (
            minpos if vmin <= 0 else vmin,
            minpos if vmax <= 0 else vmax,
        )

    def get_transform(self):
        """Get the transform."""
        return InverseScale.InverseTransform()

    class InverseTransform(transforms.Transform):
        """The transform to invert the scaling on the axis."""

        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = True

        def inverted(self):
            """Get the inverse transform."""
            return InverseScale.InverseTransform()

        def transform_non_affine(self, a):
            """Transform data."""
            # Don't spam about invalid divide by zero errors
            with numpy.errstate(divide="ignore", invalid="ignore"):
                return 1 / numpy.array(a)


class CustomScaleLocator(ticker.MaxNLocator):
    """Dynamically find tick positions on custom scales."""

    def __init__(self, is_minor=False):
        self.is_minor = is_minor

    @property
    def numticks(self):
        """Get number of ticks."""
        if self.axis is not None:
            numticks = max(1, self.axis.get_tick_space() - 4)
            # Amount of ticks is set between 3 and 9
            self._numticks = numpy.clip(numticks, 3, 9)
        else:
            self._numticks = 9
        if self.is_minor:
            # Amount of minor ticks is equal to amount of major ticks
            # times (N+1) minus N. Where N is the amount of minor ticks
            # in between the major ticks.
            self._numticks = len(self.axis.get_majorticklocs()) * 4 - 3
        return self._numticks

    @numticks.setter
    def numticks(self, numticks):
        """Set number of ticks."""
        self._numticks = numticks

    def tick_values(self, vmin, vmax):
        """Get tick values."""
        vmin, vmax = transforms.nonsingular(vmin, vmax, expander=0.05)
        vmin, vmax = min(vmin, vmax), max(vmin, vmax)  # Swap values if needed
        lin_tick_pos = numpy.linspace(vmin, vmax, self.numticks)
        lin_tick_pos = lin_tick_pos[lin_tick_pos != 0]  # Remove zeroes
        if self.axis.get_scale() == "squareroot":
            tick_pos = lin_tick_pos**2
        elif self.axis.get_scale() == "inverse":
            tick_pos = 1 / lin_tick_pos
        else:
            raise ValueError("Wrong locator for the axis type")
        tick_pos = tick_pos * ((vmax - vmin) / (max(tick_pos) - min(tick_pos)))
        tick_pos *= 2 if self.axis.get_scale() == "squareroot" else 1
        return tick_pos


class RadianLocator(ticker.MultipleLocator):
    """
    Dynamically place tick positions on radian scale.

    Places ticks at a distance of pi if there's between 4 and 8 ticks
    Otherwise it places ticks at a distance of 2pi if reasonable, or with
    a multiple of 5 pi such that a number between 3 and 8 ticks are placed
    At smaller values, the distances between the ticks are a power of 2,
    multiplied by pi. e.g. (1/2)pi, (1/4)pi, (1/8)pi etc..
    """

    def __init__(self):
        super().__init__(base=self.base)

    def tick_values(self, vmin, vmax):
        """Get tick values."""
        if vmax < vmin:
            vmin, vmax = vmax, vmin

        self._edge = ticker._Edge_integer(self.base, 0)
        step = self._edge.step
        vmin -= self._offset
        vmax -= self._offset
        vmin = self._edge.ge(vmin) * step
        n = (vmax - vmin + 0.001 * step) // step
        locs = vmin - step + numpy.arange(n + 3) * step + self._offset
        return self.raise_if_exceeds(locs)

    @property
    def base(self):
        """Get base."""
        if self.axis is None:
            return numpy.pi

        vmin, vmax = self.axis.get_view_interval()
        distance = numpy.pi
        # Amount of ticks if we use a multiple of pi
        num_ticks = (vmax - vmin) / distance
        # Desired amount of ticks, should be between 3 and 8
        numticks_goal = max(1, self.axis.get_tick_space() - 4)
        numticks_goal = numpy.clip(numticks_goal, 3, 7)
        ratio = num_ticks / numticks_goal
        if num_ticks > 8:
            if ratio < 2:  # Use a distance of 2pi if reasonable
                return distance * 2
            # Make sure ratio is never rounded to 0:
            ratio = 5 if round(ratio / 5) == 0 else ratio
            return distance * round(ratio / 5) * 5
        if num_ticks < 4:
            ratio = num_ticks / numticks_goal
            exponent = int(numpy.log2(abs(ratio)))
            result = 2**exponent  # Return distance as a power of 2
            return distance * result
        return distance


def register_scales():
    """Register custom scales."""
    scale.register_scale(Log2Scale)
    scale.register_scale(RadiansScale)
    scale.register_scale(SquareRootScale)
    scale.register_scale(InverseScale)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Wrapper classes for mpl artists.

Provides GObject based wrappers for mpl artists.
"""
from itertools import islice

from gi.repository import GObject, Graphs

from graphs import misc, utilities

from matplotlib import artist, pyplot
from matplotlib.figure import Figure

import numpy

from scipy.stats import median_abs_deviation

import sympy
from sympy.calculus.singularities import singularities as find_singularities


def new_for_item(fig: Figure, item: Graphs.Item) -> GObject.Object:
    """
    Create a new artist for an item.

    Creates bindings between item and artist properties so changes are handled
    automatically.
    """
    if isinstance(item, Graphs.DataItem):
        cls = DataItemArtistWrapper
    elif isinstance(item, Graphs.EquationItem):
        cls = EquationItemArtistWrapper
    elif isinstance(item, Graphs.TextItem):
        cls = TextItemArtistWrapper
    elif isinstance(item, Graphs.FillItem):
        cls = FillItemArtistWrapper
    artist_wrapper = cls(
        fig.axes[item.get_yposition() * 2 + item.get_xposition()],
        item,
    )
    for prop in dir(artist_wrapper.props):
        if not (prop == "label" and artist_wrapper.legend):
            item.bind_property(prop, artist_wrapper, prop, 0)
    artist_wrapper.connect("notify", lambda _x, _y: fig.update_legend())
    return artist_wrapper


class ItemArtistWrapper(GObject.Object):
    """Wrapper for base Item."""

    __gtype_name__ = "GraphsItemArtistWrapper"
    legend = False

    def get_artist(self) -> artist:
        """Get underlying mpl artist."""
        return self._artist

    @GObject.Property(type=str, default="")
    def name(self) -> str:
        """Get name/label property."""
        return self._artist.get_label()

    @name.setter
    def name(self, name: str) -> None:
        """Set name/label property."""
        self._artist.set_label(name)

    @GObject.Property(type=str, default="000000")
    def color(self) -> str:
        """Get color property."""
        return self._color_artist.get_color()

    @color.setter
    def color(self, color: str) -> None:
        """Set color property."""
        self._color_artist.set_color(color)

    @GObject.Property(type=float, default=1)
    def alpha(self) -> float:
        """Get alpha property."""
        return self._color_artist.get_alpha()

    @alpha.setter
    def alpha(self, alpha: float) -> None:
        """Set alpha property."""
        self._artist.set_alpha(alpha)


class DataItemArtistWrapper(ItemArtistWrapper):
    """Wrapper for DataItem."""

    __gtype_name__ = "GraphsDataItemArtistWrapper"
    selected = GObject.Property(type=bool, default=True)
    linewidth = GObject.Property(type=float, default=3)
    markersize = GObject.Property(type=float, default=7)
    legend = True

    @GObject.Property(type=Graphs.DataHolder)
    def data(self) -> Graphs.DataHolder:
        """Get data property."""
        raise NotImplementedError

    @data.setter
    def data(self, data: Graphs.DataHolder) -> None:
        """Set data property."""
        xdata, ydata, xerr, yerr = self._handle_singularities(data)
        self._data.set_data((xdata, ydata))

        if xerr is not None:
            start = numpy.column_stack((xdata - xerr, ydata))
            end = numpy.column_stack((xdata + xerr, ydata))
            self._xbar.set_segments(numpy.stack((start, end), axis=1))
            self._xcaps[0].set_data(xdata - xerr, ydata)
            self._xcaps[1].set_data(xdata + xerr, ydata)

        if yerr is not None:
            start = numpy.column_stack((xdata, ydata - yerr))
            end = numpy.column_stack((xdata, ydata + yerr))
            self._ybar.set_segments(numpy.stack((start, end), axis=1))
            self._ycaps[0].set_data(xdata, ydata - yerr)
            self._ycaps[1].set_data(xdata, ydata + yerr)

    @GObject.Property(type=bool, default=True)
    def showxerr(self) -> bool:
        """Get showxerr property."""
        raise NotImplementedError

    @showxerr.setter
    def showxerr(self, showxerr: bool) -> None:
        """Set showxerr property."""
        self._xbar.set_visible(showxerr)
        for cap in self._xcaps:
            cap.set_visible(showxerr)

    @GObject.Property(type=bool, default=True)
    def showyerr(self) -> bool:
        """Get showyerr property."""
        raise NotImplementedError

    @showyerr.setter
    def showyerr(self, showyerr: bool) -> None:
        """Set showyerr property."""
        self._ybar.set_visible(showyerr)
        for cap in self._ycaps:
            cap.set_visible(showyerr)

    @GObject.Property(type=int, default=1)
    def linestyle(self) -> int:
        """Get linestyle property."""
        return misc.LINESTYLES.index(self._data.get_linestyle())

    @linestyle.setter
    def linestyle(self, linestyle: int) -> None:
        """Set linestyle property."""
        self._data.set_linestyle(misc.LINESTYLES[linestyle])

    @GObject.Property(type=int, default=1)
    def markerstyle(self) -> int:
        """Get markerstyle property."""
        return misc.MARKERSTYLES.index(self._data.get_marker())

    @markerstyle.setter
    def markerstyle(self, markerstyle: int) -> None:
        """Set markerstyle property."""
        self._data.set_marker(misc.MARKERSTYLES[markerstyle])

    @GObject.Property(type=float, default=0)
    def errcapsize(self) -> float:
        """Get errcapsize property."""
        raise NotImplementedError

    @errcapsize.setter
    def errcapsize(self, errcapsize: float) -> None:
        """Set errcapsize property."""
        for cap in self._caps:
            cap.set_markersize(errcapsize * 2)

    @GObject.Property(type=float, default=1)
    def errcapthick(self) -> float:
        """Get errcapthick property."""
        raise NotImplementedError

    @errcapthick.setter
    def errcapthick(self, errcapthick: float) -> None:
        """Set errcapthick property."""
        for cap in self._caps:
            cap.set_markeredgewidth(errcapthick)

    @GObject.Property(type=float, default=1)
    def errlinewidth(self) -> float:
        """Get errlinewidth property."""
        raise NotImplementedError

    @errlinewidth.setter
    def errlinewidth(self, errlinewidth: float) -> None:
        """Set errlinewidth property."""
        for bar in self._bars:
            bar.set_linewidth(errlinewidth)

    @GObject.Property(type=bool, default=False)
    def errbarsabove(self) -> bool:
        """Get errbarsabove property."""
        raise NotImplementedError

    @errbarsabove.setter
    def errbarsabove(self, errbarsabove: bool) -> None:
        """Set errbarsabove property."""
        zorder = self._data.get_zorder()
        offset = 1 if errbarsabove else -1
        for bar in self._bars:
            bar.set_zorder(zorder + offset)

    @GObject.Property(type=str, default="")
    def errcolor(self) -> str:
        """Get errcolor property."""
        raise NotImplementedError

    @errcolor.setter
    def errcolor(self, errcolor: str) -> None:
        """Set errcolor property."""
        if not errcolor:
            return

        for bar in self._bars:
            bar.set_color(errcolor)
        for cap in self._caps:
            cap.set_color(errcolor)
            cap.set_markerfacecolor(errcolor)
            cap.set_markeredgecolor(errcolor)

    def _set_properties(self, *_args) -> None:
        linewidth, markersize = self.props.linewidth, self.props.markersize
        if not self.props.selected:
            linewidth *= 0.35
            markersize *= 0.35
        self._data.set_linewidth(linewidth)
        self._data.set_markersize(markersize)

    @staticmethod
    def _handle_singularities(data: Graphs.DataHolder) -> tuple:
        """Adjust data to handle singularity jumps."""
        xdata = utilities.bytes_to_ndarray(data.get_xdata_b())
        ydata = utilities.bytes_to_ndarray(data.get_ydata_b())
        xerr = utilities.bytes_to_ndarray(data.get_xerr_b())
        yerr = utilities.bytes_to_ndarray(data.get_yerr_b())

        if len(xdata) < 2:
            return xdata, ydata, xerr, yerr

        # Detect singularities using Median Absolute Deviation
        grad = numpy.abs(numpy.gradient(ydata, xdata))
        median = numpy.median(grad)
        mad = median_abs_deviation(grad, scale="normal")

        if mad == 0:
            mad = (xdata[1] - xdata[0]) * 0.01

        threshold = median + 6 * mad
        sign_change = numpy.sign(ydata[:-1]) != numpy.sign(ydata[1:])
        mask = (grad[:-1] > threshold) & sign_change

        if not numpy.any(mask):
            return xdata, ydata, xerr, yerr

        edges = numpy.diff(mask.astype(int))
        starts = numpy.where(edges == 1)[0] + 1
        ends = numpy.where(edges == -1)[0] + 1

        if mask[0]:
            starts = numpy.r_[0, starts]
        if mask[-1]:
            ends = numpy.r_[ends, len(mask)]

        mask = numpy.zeros_like(mask, dtype=bool)
        mask[(starts + ends) // 2] = True

        bad_points = numpy.zeros(len(xdata), dtype=bool)
        left = numpy.abs(ydata[:-1]) > numpy.abs(ydata[1:])
        bad_points[:-1] |= mask & left
        bad_points[1:] |= mask & ~left

        xdata = xdata.copy()
        ydata = ydata.copy()
        xdata[bad_points] = numpy.nan
        ydata[bad_points] = numpy.nan

        if xerr is not None:
            xerr = xerr.copy()
            xerr[bad_points] = numpy.nan

        if yerr is not None:
            yerr = yerr.copy()
            yerr[bad_points] = numpy.nan

        return xdata, ydata, xerr, yerr

    def __init__(self, axis: pyplot.axis, item: Graphs.Item) -> None:
        super().__init__()
        xdata, ydata, xerr, yerr = self._handle_singularities(item.props.data)
        self._artist = axis.errorbar(
            xdata,
            ydata,
            xerr=xerr,
            yerr=yerr,
            label=item.get_name(),
            color=item.get_color(),
            alpha=item.get_alpha(),
            linestyle=misc.LINESTYLES[item.get_linestyle()],
            marker=misc.MARKERSTYLES[item.get_markerstyle()],
            capsize=item.get_errcapsize(),
            capthick=item.get_errcapthick(),
            elinewidth=item.get_errlinewidth(),
            barsabove=item.get_errbarsabove(),
            ecolor=item.get_errcolor(),
        )

        self._data, self._caps, self._bars = self._artist
        self._color_artist = self._data

        # We iterate over bar and caps in assignments to handle all
        # combinations with error bars on either or both axes.
        bar_iter = iter(self._bars)
        cap_iter = iter(self._caps)

        if xerr is not None:
            self._xbar = next(bar_iter)
            self._xcaps = tuple(islice(cap_iter, 2))
            if not item.get_showxerr():
                self._xbar.set_visible(False)
                for cap in self._xcaps:
                    cap.set_visible(False)
        if yerr is not None:
            self._ybar = next(bar_iter)
            self._ycaps = tuple(islice(cap_iter, 2))
            if not item.get_showyerr():
                self._ybar.set_visible(False)
                for cap in self._ycaps:
                    cap.set_visible(False)

        for prop in ("selected", "linewidth", "markersize"):
            self.set_property(prop, item.get_property(prop))
            self.connect(f"notify::{prop}", self._set_properties)
        self._set_properties()


class EquationItemArtistWrapper(ItemArtistWrapper):
    """Wrapper for EquationItem."""

    __gtype_name__ = "GraphsEquationItemArtistWrapper"
    selected = GObject.Property(type=bool, default=True)
    linewidth = GObject.Property(type=float, default=3)
    legend = True
    _singularities_cache = {}

    def __init__(self, axis: pyplot.axis, item: Graphs.Item):
        super().__init__()

        equation = item.get_preprocessed_equation()
        self._equation = equation
        self._expr = sympy.sympify(equation)
        self._axis = axis
        self._view_change_timeout_id = None
        axis.callbacks.connect("xlim_changed", self._on_view_change)
        self._artist = axis.plot(
            [],
            [],
            label=item.get_name(),
            color=item.get_color(),
            alpha=item.get_alpha(),
            linestyle=misc.LINESTYLES[item.get_linestyle() + 1],
            marker="none",
        )[0]
        self._color_artist = self._artist
        for prop in ("selected", "linewidth"):
            self.set_property(prop, item.get_property(prop))
            self.connect(f"notify::{prop}", self._set_properties)
        self._set_properties(None, None)
        self._generate_data()

    def _timeout_callback(self) -> bool:
        self._view_change_timeout_id = None
        self._generate_data()
        return False

    def _on_view_change(self, *_args) -> None:
        """Debounced view change handler that generates data after delay."""
        if self._view_change_timeout_id is not None:
            GObject.source_remove(self._view_change_timeout_id)
        self._view_change_timeout_id = \
            GObject.timeout_add(100, self._timeout_callback)

    @GObject.Property(type=str, flags=2)
    def equation(self) -> None:
        """Write-only property, ignored."""

    @equation.setter
    def equation(self, equation: str) -> None:
        self._singularities_cache.clear()
        self._equation = Graphs.preprocess_equation(equation)
        self._expr = sympy.sympify(self._equation)
        self._generate_data()

    @GObject.Property(type=int, default=1)
    def linestyle(self) -> int:
        """Get linestyle property."""
        return misc.LINESTYLES.index(self._artist.get_linestyle()) - 1

    @linestyle.setter
    def linestyle(self, linestyle: int) -> None:
        """Set linestyle property."""
        self._artist.set_linestyle(misc.LINESTYLES[linestyle + 1])

    def _set_properties(self, _x, _y) -> None:
        linewidth = self.props.linewidth
        if not self.props.selected:
            linewidth *= 0.35
        self._artist.set_linewidth(linewidth)

    def _generate_data(self):
        """Generate new data for the artist."""
        x_start, x_stop = self._axis.get_xlim()
        scale = Graphs.scale_from_string(self._axis.get_xscale())

        lower = Graphs.get_value_at_fraction(-1, x_start, x_stop, scale)
        upper = Graphs.get_value_at_fraction(2, x_start, x_stop, scale)
        limits = (lower, upper)

        data = utilities.equation_to_data(self._equation, limits, scale=scale)
        singularities = self._find_singularities(limits)
        if singularities:
            data = self._insert_singularity_points(data, singularities)

        self._artist.set_data(*data)
        if self._axis.figure.parent is not None:
            self._axis.figure.parent.queue_draw()

    def _find_singularities(self, limits):
        lower, upper = limits

        cached = self._singularities_cache.get(self._equation)
        if cached:
            cached_min, cached_max = cached["limits"]

            if lower >= cached_min and upper <= cached_max:
                return {
                    s
                    for s in cached["singularities"] if lower <= s <= upper
                }

            x_min, x_max = min(lower, cached_min), max(upper, cached_max)
        else:
            x_min, x_max = lower, upper

        domain = sympy.Interval(x_min, x_max)
        all_singularities = find_singularities(self._expr, misc.X, domain)

        self._singularities_cache[self._equation] = {
            "limits": (x_min, x_max),
            "singularities": all_singularities,
        }

        return {s for s in all_singularities if lower <= s <= upper}

    def _insert_singularity_points(self, data, singularities) -> tuple:
        """Insert NaN and infinite value points at singularities."""
        xdata, ydata = data
        singularities_arr = numpy.fromiter(sorted(singularities), dtype=float)
        indices = numpy.searchsorted(xdata, singularities_arr)

        n = len(xdata)
        triple_mask = (indices > 1) & (indices < n - 1)
        new_size = n + len(indices) + 2 * triple_mask.sum()

        ylim = self._axis.get_ylim()
        ylim_range = abs(ylim[1] - ylim[0])
        ydata_range = numpy.nanmax(ydata) - numpy.nanmin(ydata)
        inf_value = max(ylim_range * 1.5, ydata_range * 1.5) * 2
        epsilon = (xdata[1] - xdata[0]) * 0.01

        # shift indices due to previous insertions
        shifts = numpy.cumsum(1 + 2 * triple_mask) - (1 + 2 * triple_mask)
        target_indices = indices + shifts

        triple_idxs = indices[triple_mask]
        triple_targets = target_indices[triple_mask]
        triple_values = singularities_arr[triple_mask]

        data_mask = numpy.ones(new_size, dtype=bool)
        data_mask[target_indices] = False
        data_mask[triple_targets + 1] = False
        data_mask[triple_targets + 2] = False

        x_new = numpy.empty(new_size, dtype=float)
        y_new = numpy.empty(new_size, dtype=float)
        x_new[data_mask] = xdata
        y_new[data_mask] = ydata

        left = numpy.sign(ydata[triple_idxs - 1] - ydata[triple_idxs - 2])
        right = -numpy.sign(ydata[triple_idxs + 1] - ydata[triple_idxs])
        inf_values = inf_value + ydata[triple_idxs]

        x_new[triple_targets] = triple_values - epsilon
        x_new[triple_targets + 1] = numpy.nan
        x_new[triple_targets + 2] = triple_values + epsilon

        y_new[triple_targets] = left * inf_values
        y_new[triple_targets + 1] = numpy.nan
        y_new[triple_targets + 2] = right * inf_values

        single_targets = target_indices[~triple_mask]
        x_new[single_targets] = numpy.nan
        y_new[single_targets] = numpy.nan

        return x_new, y_new


class TextItemArtistWrapper(ItemArtistWrapper):
    """Wrapper for TextItem."""

    __gtype_name__ = "GraphsTextItemArtistWrapper"

    @GObject.Property(type=float, default=12)
    def size(self) -> float:
        """Get size property."""
        return self._artist.get_fontsize()

    @size.setter
    def size(self, size: float) -> None:
        """Set size property."""
        self._artist.set_fontsize(size)

    @GObject.Property(type=int, default=0, minimum=0, maximum=360)
    def rotation(self) -> int:
        """Get rotation property."""
        return self._artist.get_rotation()

    @rotation.setter
    def rotation(self, rotation: int) -> None:
        """Set rotation property."""
        self._artist.set_rotation(rotation)

    @GObject.Property(type=str, default="")
    def text(self) -> str:
        """Get text property."""
        return self._artist.get_text()

    @text.setter
    def text(self, text: str) -> None:
        """Set text property."""
        self._artist.set_text(text)

    @GObject.Property(type=float, default=0)
    def xanchor(self) -> float:
        """Get xanchor property."""
        return self._artist.get_position()[0]

    @xanchor.setter
    def xanchor(self, xanchor: float) -> None:
        """Set xanchor property."""
        self._artist.set_position((xanchor, self.props.yanchor))

    @GObject.Property(type=float, default=0)
    def yanchor(self) -> float:
        """Get yanchor property."""
        return self._artist.get_position()[1]

    @yanchor.setter
    def yanchor(self, yanchor: float) -> None:
        """Set yanchor property."""
        self._artist.set_position((self.props.xanchor, yanchor))

    def __init__(self, axis: pyplot.axis, item: Graphs.Item):
        super().__init__()
        self._artist = axis.text(
            item.get_xanchor(),
            item.get_yanchor(),
            item.get_text(),
            label=item.get_name(),
            color=item.get_color(),
            alpha=item.get_alpha(),
            fontsize=item.get_size(),
            rotation=item.get_rotation(),
            clip_on=True,
        )
        self._color_artist = self._artist


class FillItemArtistWrapper(ItemArtistWrapper):
    """Wrapper for FillItem."""

    __gtype_name__ = "GraphsFillItemArtistWrapper"

    @GObject.Property(type=object, flags=2)
    def data(self) -> None:
        """Write-only property, ignored."""

    @data.setter
    def data(self, data) -> None:
        dummy = Figure().add_subplot().fill_between(*data)
        self._artist.set_paths([dummy.get_paths()[0].vertices])

    def __init__(self, axis: pyplot.axis, item: Graphs.Item):
        super().__init__()
        self._artist = axis.fill_between(
            *item.get_data_tuple(),
            label=item.get_name(),
            color=item.get_color(),
            alpha=item.get_alpha(),
        )
        self._color_artist = self._artist
